@@ -783,7 +783,8 @@ int connman_inet_del_ipv6_network_route(int index, const char *host,
 	close(sk);
 out:
 	if (err < 0)
-		connman_error("Del IPv6 host route error");
+		connman_error("Del IPv6 host route error (%s)",
+						strerror(errno));
 
 	return err;
 }
@@ -795,7 +796,7 @@ int connman_inet_del_ipv6_host_route(int index, const char *host)
 
 int connman_inet_add_ipv6_network_route(int index, const char *host,
 					const char *gateway,
-						unsigned char prefix_len)
+					unsigned char prefix_len)
 {
 	struct in6_rtmsg rt;
 	int sk, err;
@@ -833,7 +834,8 @@ int connman_inet_add_ipv6_network_route(int index, const char *host,
 	close(sk);
 out:
 	if (err < 0)
-		connman_error("Set IPv6 host route error");
+		connman_error("Set IPv6 host route error (%s)",
+						strerror(errno));
 
 	return err;
 }
@@ -875,7 +877,8 @@ int connman_inet_set_ipv6_gateway_address(int index, const char *gateway)
 	close(sk);
 out:
 	if (err < 0)
-		connman_error("Set default IPv6 gateway error");
+		connman_error("Set default IPv6 gateway error (%s)",
+						strerror(errno));
 
 	return err;
 }
@@ -911,7 +914,8 @@ int connman_inet_clear_ipv6_gateway_address(int index, const char *gateway)
 	close(sk);
 out:
 	if (err < 0)
-		connman_error("Clear default IPv6 gateway error");
+		connman_error("Clear default IPv6 gateway error (%s)",
+						strerror(errno));
 
 	return err;
 }
@@ -1010,6 +1014,52 @@ int connman_inet_set_gateway_interface(int index)
 	return err;
 }
 
+int connman_inet_set_ipv6_gateway_interface(int index)
+{
+	struct ifreq ifr;
+	struct rtentry rt;
+	struct sockaddr_in6 addr;
+	const struct in6_addr any = IN6ADDR_ANY_INIT;
+	int sk, err;
+
+	DBG("");
+
+	sk = socket(PF_INET6, SOCK_DGRAM, 0);
+	if (sk < 0)
+		return -1;
+
+	memset(&ifr, 0, sizeof(ifr));
+	ifr.ifr_ifindex = index;
+
+	if (ioctl(sk, SIOCGIFNAME, &ifr) < 0) {
+		close(sk);
+		return -1;
+	}
+
+	DBG("ifname %s", ifr.ifr_name);
+
+	memset(&rt, 0, sizeof(rt));
+	rt.rt_flags = RTF_UP;
+
+	memset(&addr, 0, sizeof(addr));
+	addr.sin6_family = AF_INET6;
+	addr.sin6_addr = any;
+
+	memcpy(&rt.rt_genmask, &addr, sizeof(rt.rt_genmask));
+	memcpy(&rt.rt_dst, &addr, sizeof(rt.rt_dst));
+	memcpy(&rt.rt_gateway, &addr, sizeof(rt.rt_gateway));
+
+	rt.rt_dev = ifr.ifr_name;
+
+	err = ioctl(sk, SIOCADDRT, &rt);
+	if (err < 0)
+		connman_error("Setting default interface route failed (%s)",
+							strerror(errno));
+	close(sk);
+
+	return err;
+}
+
 int connman_inet_clear_gateway_address(int index, const char *gateway)
 {
 	struct ifreq ifr;
@@ -1106,6 +1156,52 @@ int connman_inet_clear_gateway_interface(int index)
 	return err;
 }
 
+int connman_inet_clear_ipv6_gateway_interface(int index)
+{
+	struct ifreq ifr;
+	struct rtentry rt;
+	struct sockaddr_in6 addr;
+	const struct in6_addr any = IN6ADDR_ANY_INIT;
+	int sk, err;
+
+	DBG("");
+
+	sk = socket(PF_INET6, SOCK_DGRAM, 0);
+	if (sk < 0)
+		return -1;
+
+	memset(&ifr, 0, sizeof(ifr));
+	ifr.ifr_ifindex = index;
+
+	if (ioctl(sk, SIOCGIFNAME, &ifr) < 0) {
+		close(sk);
+		return -1;
+	}
+
+	DBG("ifname %s", ifr.ifr_name);
+
+	memset(&rt, 0, sizeof(rt));
+	rt.rt_flags = RTF_UP;
+
+	memset(&addr, 0, sizeof(addr));
+	addr.sin6_family = AF_INET6;
+	addr.sin6_addr = any;
+
+	memcpy(&rt.rt_genmask, &addr, sizeof(rt.rt_genmask));
+	memcpy(&rt.rt_dst, &addr, sizeof(rt.rt_dst));
+	memcpy(&rt.rt_gateway, &addr, sizeof(rt.rt_gateway));
+
+	rt.rt_dev = ifr.ifr_name;
+
+	err = ioctl(sk, SIOCDELRT, &rt);
+	if (err < 0)
+		connman_error("Removing default interface route failed (%s)",
+							strerror(errno));
+	close(sk);
+
+	return err;
+}
+
 connman_bool_t connman_inet_compare_subnet(int index, const char *host)
 {
 	struct ifreq ifr;
