@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -23,8 +23,14 @@
 #include <config.h>
 #endif
 
+#include <errno.h>
+#include <unistd.h>
+
 #include "connman.h"
 
+#define PROFILE_SUFFIX	"profile"
+#define CONFIG_SUFFIX	"config"
+
 static GSList *storage_list = NULL;
 
 static gint compare_priority(gconstpointer a, gconstpointer b)
@@ -66,17 +72,16 @@ void connman_storage_unregister(struct connman_storage *storage)
 	storage_list = g_slist_remove(storage_list, storage);
 }
 
-GKeyFile *__connman_storage_open(void)
+GKeyFile *__connman_storage_open(const char *ident, const char *suffix)
 {
 	GKeyFile *keyfile;
 	gchar *pathname, *data = NULL;
 	gboolean result;
 	gsize length;
 
-	DBG("");
+	DBG("ident %s suffix %s", ident, suffix);
 
-	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR,
-					__connman_profile_active_ident());
+	pathname = g_strdup_printf("%s/%s.%s", STORAGEDIR, ident, suffix);
 	if (pathname == NULL)
 		return NULL;
 
@@ -84,39 +89,37 @@ GKeyFile *__connman_storage_open(void)
 
 	g_free(pathname);
 
-	if (result == FALSE)
-		return NULL;
-
 	keyfile = g_key_file_new();
 
-	if (length > 0) {
-		if (g_key_file_load_from_data(keyfile, data, length,
-							0, NULL) == FALSE)
-			goto done;
-	}
+	if (result == FALSE)
+		goto done;
+
+	if (length > 0)
+		g_key_file_load_from_data(keyfile, data, length, 0, NULL);
 
-done:
 	g_free(data);
 
+done:
 	DBG("keyfile %p", keyfile);
 
 	return keyfile;
 }
 
-void __connman_storage_close(GKeyFile *keyfile, gboolean save)
+void __connman_storage_close(const char *ident, const char *suffix,
+					GKeyFile *keyfile, gboolean save)
 {
 	gchar *pathname, *data = NULL;
 	gsize length = 0;
 
-	DBG("keyfile %p save %d", keyfile, save);
+	DBG("ident %s suffix %s keyfile %p save %d",
+					ident, suffix, keyfile, save);
 
 	if (save == FALSE) {
 		g_key_file_free(keyfile);
 		return;
 	}
 
-	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR,
-					__connman_profile_active_ident());
+	pathname = g_strdup_printf("%s/%s.%s", STORAGEDIR, ident, suffix);
 	if (pathname == NULL)
 		return;
 
@@ -132,7 +135,53 @@ void __connman_storage_close(GKeyFile *keyfile, gboolean save)
 	g_key_file_free(keyfile);
 }
 
-int __connman_storage_load_global(void)
+void __connman_storage_delete(const char *ident, const char *suffix)
+{
+	gchar *pathname;
+
+	DBG("ident %s suffix %s", ident, suffix);
+
+	pathname = g_strdup_printf("%s/%s.%s", STORAGEDIR, ident, suffix);
+	if (pathname == NULL)
+		return;
+
+	if (unlink(pathname) < 0)
+		connman_error("Failed to remove %s", pathname);
+}
+
+GKeyFile *__connman_storage_open_profile(const char *ident)
+{
+	return __connman_storage_open(ident, PROFILE_SUFFIX);
+}
+
+void __connman_storage_close_profile(const char *ident,
+					GKeyFile *keyfile, gboolean save)
+{
+	__connman_storage_close(ident, PROFILE_SUFFIX, keyfile, save);
+}
+
+void __connman_storage_delete_profile(const char *ident)
+{
+	__connman_storage_delete(ident, PROFILE_SUFFIX);
+}
+
+GKeyFile *__connman_storage_open_config(const char *ident)
+{
+	return __connman_storage_open(ident, CONFIG_SUFFIX);
+}
+
+void __connman_storage_close_config(const char *ident,
+					GKeyFile *keyfile, gboolean save)
+{
+	__connman_storage_close(ident, CONFIG_SUFFIX, keyfile, save);
+}
+
+void __connman_storage_delete_config(const char *ident)
+{
+	__connman_storage_delete(ident, CONFIG_SUFFIX);
+}
+
+int __connman_storage_init_profile(void)
 {
 	GSList *list;
 
@@ -141,8 +190,8 @@ int __connman_storage_load_global(void)
 	for (list = storage_list; list; list = list->next) {
 		struct connman_storage *storage = list->data;
 
-		if (storage->global_load) {
-			if (storage->global_load() == 0)
+		if (storage->profile_init) {
+			if (storage->profile_init() == 0)
 				return 0;
 		}
 	}
@@ -150,17 +199,17 @@ int __connman_storage_load_global(void)
 	return -ENOENT;
 }
 
-int __connman_storage_save_global(void)
+int __connman_storage_load_profile(struct connman_profile *profile)
 {
 	GSList *list;
 
-	DBG("");
+	DBG("profile %p", profile);
 
 	for (list = storage_list; list; list = list->next) {
 		struct connman_storage *storage = list->data;
 
-		if (storage->global_save) {
-			if (storage->global_save() == 0)
+		if (storage->profile_load) {
+			if (storage->profile_load(profile) == 0)
 				return 0;
 		}
 	}
@@ -168,35 +217,17 @@ int __connman_storage_save_global(void)
 	return -ENOENT;
 }
 
-int __connman_storage_load_device(struct connman_device *device)
+int __connman_storage_save_profile(struct connman_profile *profile)
 {
 	GSList *list;
 
-	DBG("device %p", device);
+	DBG("profile %p", profile);
 
 	for (list = storage_list; list; list = list->next) {
 		struct connman_storage *storage = list->data;
 
-		if (storage->device_load) {
-			if (storage->device_load(device) == 0)
-				return 0;
-		}
-	}
-
-	return -ENOENT;
-}
-
-int __connman_storage_save_device(struct connman_device *device)
-{
-	GSList *list;
-
-	DBG("device %p", device);
-
-	for (list = storage_list; list; list = list->next) {
-		struct connman_storage *storage = list->data;
-
-		if (storage->device_save) {
-			if (storage->device_save(device) == 0)
+		if (storage->profile_save) {
+			if (storage->profile_save(profile) == 0)
 				return 0;
 		}
 	}
@@ -240,6 +271,42 @@ int __connman_storage_save_service(struct connman_service *service)
 	return -ENOENT;
 }
 
+int __connman_storage_load_device(struct connman_device *device)
+{
+	GSList *list;
+
+	DBG("device %p", device);
+
+	for (list = storage_list; list; list = list->next) {
+		struct connman_storage *storage = list->data;
+
+		if (storage->device_load) {
+			if (storage->device_load(device) == 0)
+				return 0;
+		}
+	}
+
+	return -ENOENT;
+}
+
+int __connman_storage_save_device(struct connman_device *device)
+{
+	GSList *list;
+
+	DBG("device %p", device);
+
+	for (list = storage_list; list; list = list->next) {
+		struct connman_storage *storage = list->data;
+
+		if (storage->device_save) {
+			if (storage->device_save(device) == 0)
+				return 0;
+		}
+	}
+
+	return -ENOENT;
+}
+
 int __connman_storage_init(void)
 {
 	DBG("");
